import functools
from pathlib import Path
from typing import List

from nymbus.config.components.component import Component
from nymbus.config.environment import DEFAULT_ENVIRONMENT, DEFAULT_ENVIRONMENT_FILE, DEFAULT_NYMBUS_EXTENSION
from nymbus.config.envspec import EnvSpec
from nymbus.config.readers.reader import merge_dictionaries, read_yml
from nymbus.config.components.step import Step


def read_dicts_merging(context: Path, environment: str = DEFAULT_ENVIRONMENT, optional_env: bool = True) -> dict:
    # Read the default
    default_location = Path(context) / DEFAULT_ENVIRONMENT_FILE
    default = read_yml(default_location)

    # Read the component from the location
    location = Path(context) / f"{environment}{DEFAULT_NYMBUS_EXTENSION}"
    if location.exists() or not optional_env:
        specific = read_yml(location)

        # Merge the dictionaries
        result = merge_dictionaries(default, specific)
    else:
        result = default

    return result


def read_component(context: Path, environment: str = DEFAULT_ENVIRONMENT):
    result = read_dicts_merging(context, environment, optional_env=False)
    return Component(context.name, environment, result)


def _has_env_spec(location: Path) -> bool:
    return (location/DEFAULT_ENVIRONMENT_FILE).exists()


def find_env_spec_locations(location: Path, include_current: bool = False) -> List[Path]:

    # Go up the uppermost env file
    env_specs = []
    while _has_env_spec(location):
        env_specs.insert(0, location)
        location = location.parent

    # Keep the current (i.e. the last one)?
    return env_specs[:-1 if not include_current else len(env_specs)]


def read_upper_env_specs_merging(location: Path, environment: str = DEFAULT_ENVIRONMENT, optional_env: bool = True) -> EnvSpec:

    # Find all upper EnvSpec paths
    env_spec_paths = find_env_spec_locations(location)

    # Read them all
    env_specs = [
        read_dicts_merging(location, environment, optional_env=optional_env)
        for location in env_spec_paths
    ]

    # Merge EnvSpecs
    result = functools.reduce(merge_dictionaries, env_specs)
    return EnvSpec(result)


def merge_envs(location: Path, component: Component, target_step: Step = None, environment: str = DEFAULT_ENVIRONMENT) -> EnvSpec:

    # Read the upper envs, merging them
    env_spec = read_upper_env_specs_merging(location, environment)

    # Reduce them with component and step
    return EnvSpec.from_env(functools.reduce(
        merge_dictionaries,
        [env_spec.env, component.env] + ([target_step.env] if target_step else [])
    ))





def log_title(logger, title: str, char: str = "*", ending: bool = False):
    logger.info("")
    if not ending:
        logger.info(char*60)
    logger.info(f"   {title}   ".center(60, char))
    logger.info(char*60)



import logging
import os
import re

logger = logging.getLogger(__name__)

DEFAULT_ENVIRONMENT = "default"
NYMBUS_FOLDER = ".nymbus"
NYMBUS_OUTPUT_FOLDER = f"{NYMBUS_FOLDER}/output"
DEFAULT_NYMBUS_EXTENSION = ".nymbus.yml"
DEFAULT_ENVIRONMENT_FILE = f"{DEFAULT_ENVIRONMENT}{DEFAULT_NYMBUS_EXTENSION}"
VARIABLE_EXPANSION = re.compile(r"^\$\{([^}]+)}$")


def expand_variable(value: str):

    # Check if the value is in the form ${<name>}
    match = VARIABLE_EXPANSION.match(value)
    if not match:
        return value
    else:
        # Get the value corresponding to the variable with name <name>
        variable_name = match.group(1)
        variable_value = os.environ.get(variable_name, None)
        if variable_value is None:
            logger.warning(f"Variable {variable_name} is not set.")
            variable_value = ""
        return variable_value



class Template:

    def __init__(self, yml: dict):
        # Pop the config entries
        config = yml.copy()
        self.repository = config.pop("repository")
        self.tag = config.pop("tag", "master")



from pathlib import Path

import yaml


def read_yml(path: Path) -> dict:
    # Read the bag file
    with open(path) as f:
        return yaml.load(f, Loader=yaml.SafeLoader) or {}


def merge_dictionaries(main_dict, new_dict, merging_lists=False):
    """ Merge two dictionaries prioritizing the second one """

    # Copy the dict
    main_dict = main_dict.copy()

    # Get all the keys
    main_keys = main_dict.keys()
    new_keys = new_dict.keys()
    keys = set(main_keys).union(new_keys)

    # For each key, merge the dict
    for key in keys:
        # If key is just in new, add it to main
        if key not in main_keys:
            main_dict[key] = new_dict[key]
        # If key is in both dicts, and...
        elif key in new_keys:
            # ...the values are dicts, merge them recursively
            if isinstance(main_dict[key], dict) and isinstance(new_dict[key], dict):
                main_dict[key] = merge_dictionaries(main_dict[key], new_dict[key], merging_lists=merging_lists)
            # ...the value is NOT a dict in both cases, override the main value
            else:
                # If merging lists is True, and values are lists, concatenate the elements
                if merging_lists and isinstance(main_dict[key], list) and isinstance(new_dict[key], list):
                    main_dict[key] = main_dict[key] + new_dict[key]
                # Override them otherwise
                else:
                    main_dict[key] = new_dict[key]
        # If key is just in main, nothing to do: just keep main_dict as it is

    return main_dict



import logging
import os
from pathlib import Path

from nymbus.cli.console.output import log_title
from nymbus.config.environment import DEFAULT_ENVIRONMENT
from nymbus.config.readers.combiner import read_component, merge_envs, find_env_spec_locations
from nymbus.tools import system

logger = logging.getLogger(__name__)


class Runner:

    def run(self, component_path: str, environment: str = DEFAULT_ENVIRONMENT, step_name: str = None, context: str = None) -> None:

        # Read component
        location = component_path if Path(component_path).is_absolute() else Path(os.getcwd())/component_path
        component = read_component(location, environment)

        # Check step existence
        if step_name and step_name not in component.steps:
            raise Exception(f"Step {step_name} is not in {location}")

        # Run it (or them, if step is not specified)
        steps = [step_name] if step_name else component.steps
        for target_step_name in steps:

            # Get the target step
            target_step = component.steps[target_step_name]
            log_title(logger, target_step.name)

            # Merge envs
            env_spec = merge_envs(location, component, target_step, environment)
            env_spec.expand()

            # Run the step
            context = context or component.context or location
            system.run(target_step, env_spec, location, context)



from setuptools import setup, find_packages
import os


# Read requirements
this_location = os.path.dirname(os.path.realpath(__file__))
requirements_file = os.path.join(this_location, "requirements.txt")
with open(requirements_file, "r") as f:
    requirements = f.read().splitlines()

# Read README
readme_file = os.path.join(this_location, "README.md")
with open(readme_file, "r", encoding="utf-8") as f:
    readme = f.read()


# Package configuration
_NAME = "nymbus"
setup(
    name=_NAME,
    version="0.0.1",
    description="Deploy any code anywhere",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=3.7",
    entry_points={
        'console_scripts': [
            'nymbus = nymbus.__main__:main',
        ],
    },

    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/nymbus-tools/nymbus",
    author="Federico Pugliese",
    author_email="federico.pugliese.wr@gmail.com",
    license="Apache Software License, Version 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ]
)



import logging

from nymbus.config.environment import DEFAULT_NYMBUS_EXTENSION
from nymbus.config.envspec import EnvSpec
from nymbus.config.images.image import Image
from nymbus.config.template import Template

logger = logging.getLogger(__name__)


class Step(EnvSpec):

    def __init__(self, name: str, environment: str, yml: dict):
        super().__init__(yml)
        self.name = name
        self.environment = environment

        # Pop the config entries
        config = yml.copy()
        self.command = config.pop("command", None)
        image = config.pop("image", None)
        self.image = Image(image) if image else None
        if not self.command and not self.image:
            raise RuntimeError("One of 'command' or 'image' must be not null")
        template = config.pop("template", None)
        self.template = Template(template) if template else None

        # If there are still configs, they are unknown. Print a warning (for retro-compatibility)
        config.pop("env", {})
        if config:
            logger.warning(f"Unknown configuration in component \"{self.name}\", "
                           f"file {self.environment}{DEFAULT_NYMBUS_EXTENSION}: {config}")
